#include <pthread.h>
#include <stdlib.h>
#include <string.h>

#include "lsm.h"
#include "lsm/store.h"
#include "lsm/store_internal.h"
#include "lsm/trie.h"

lsm_error lsm_store_init(lsm_store **ptr) {
  lsm_store *store = calloc(1, sizeof(lsm_store));

  if (store == NULL) {
    return lsm_error_failed_alloc;
  }

  lsm_error res = lsm_trie_init(&store->trie);

  if (res != lsm_error_ok) {
    free(store);

    return res;
  }

  pthread_mutex_init(&store->db.lock, NULL);
  pthread_mutex_init(&store->idx.lock, NULL);

  *ptr = store;

  return lsm_error_ok;
}

uint64_t lsm_store_size(const lsm_store *store) {
  return lsm_trie_size(store->trie);
}

lsm_error lsm_store_open_read(lsm_entry_handle **out, lsm_store *store,
                              lsm_str *key) {
  lsm_entry_wrapper *wrapper;

  LSM_RES(lsm_trie_search((void **)&wrapper, store->trie, key));

  // Try to get a read lock on the entry's lock
  if (pthread_rwlock_tryrdlock(&wrapper->lock) != 0) {
    return lsm_error_lock_busy;
  }

  // While the trie's data field will never be NULL, the actual entry pointer
  // might be
  if (wrapper->entry == NULL) {
    pthread_rwlock_unlock(&wrapper->lock);

    return lsm_error_not_found;
  }

  lsm_entry_handle *handle;
  lsm_error res = lsm_entry_handle_init(&handle);

  if (res != lsm_error_ok) {
    pthread_rwlock_unlock(&wrapper->lock);

    return res;
  }

  handle->wrapper = wrapper;
  handle->store = store;
  *out = handle;

  return lsm_error_ok;
}

lsm_error lsm_store_open_write(lsm_entry_handle **out, lsm_store *store,
                               lsm_str *key) {
  lsm_entry_wrapper *wrapper;

  LSM_RES(lsm_trie_search((void **)&wrapper, store->trie, key));

  // Try to get a write lock on the entry's lock
  // TODO make this timeout to not block
  if (pthread_rwlock_wrlock(&wrapper->lock) != 0) {
    return lsm_error_lock_busy;
  }

  // While the trie's data field will never be NULL, the actual entry pointer
  // might be
  if (wrapper->entry == NULL) {
    pthread_rwlock_unlock(&wrapper->lock);

    return lsm_error_not_found;
  }

  lsm_entry_handle *handle;
  lsm_error res = lsm_entry_handle_init(&handle);

  if (res != lsm_error_ok) {
    pthread_rwlock_unlock(&wrapper->lock);

    return res;
  }

  handle->wrapper = wrapper;
  handle->store = store;
  *out = handle;

  return lsm_error_ok;
}

lsm_error lsm_store_insert(lsm_entry_handle **out, lsm_store *store,
                           lsm_str *key) {
  // TODO what happens when two inserts to the same key happen at the same time?
  lsm_entry_wrapper *wrapper;

  // If a key was previously removed from the trie, the wrapper will already be
  // present in the trie
  if (lsm_trie_search((void **)&wrapper, store->trie, key) ==
      lsm_error_not_found) {
    LSM_RES(lsm_entry_wrapper_init(&wrapper));
    pthread_rwlock_wrlock(&wrapper->lock);

    lsm_error res = lsm_trie_insert(store->trie, key, wrapper);

    // Check if entry isn't already present in advance
    if (res != lsm_error_ok) {
      lsm_entry_wrapper_free(wrapper);

      return res;
    }
  } else {
    pthread_rwlock_wrlock(&wrapper->lock);

    if (wrapper->entry != NULL) {
      pthread_rwlock_unlock(&wrapper->lock);

      return lsm_error_already_present;
    }
  }

  lsm_entry *entry;
  LSM_RES(lsm_entry_init(&entry));

  entry->key = key;
  wrapper->entry = entry;

  lsm_entry_handle *handle;
  LSM_RES(lsm_entry_handle_init(&handle));

  // No need to set the handle's file, as the entry doesn't have any data yet
  handle->wrapper = wrapper;
  handle->store = store;

  // Newly inserted entries are always dirty
  handle->states |= lsm_entry_handle_state_new;

  *out = handle;

  return lsm_error_ok;
}

void lsm_entry_remove(lsm_entry_handle *handle) {
  handle->states |= lsm_entry_handle_state_removed;
}

lsm_error lsm_entry_data_append(lsm_entry_handle *handle, lsm_str *data) {
  if (lsm_str_len(data) == 0) {
    return lsm_error_ok;
  }

  lsm_entry *entry = handle->wrapper->entry;

  uint64_t new_len = entry->data_len + lsm_str_len(data);
  const char *data_s = lsm_str_ptr(data);

  // Entries don't open their file unless needed
  if (handle->f == NULL) {
    char data_path[lsm_entry_data_path_len(handle) + 1];
    lsm_entry_data_path(data_path, handle);

    FILE *f = fopen(data_path, "ab");

    if (f == NULL) {
      return lsm_error_failed_io;
    }

    handle->f = f;
  }

  size_t written = 0;

  // TODO what happens when I/O fails?
  while (written < data->len) {
    written +=
        fwrite(&data_s[written], sizeof(char), data->len - written, handle->f);
  }

  entry->data_len = new_len;
  handle->states |= lsm_entry_handle_state_updated;

  return lsm_error_ok;
}

lsm_error lsm_entry_data_read(uint64_t *out, char *buf,
                              lsm_entry_handle *handle, uint64_t len) {
  const lsm_entry *entry = handle->wrapper->entry;

  if (entry->data_len == 0) {
    *out = 0;

    return lsm_error_ok;
  }

  // Entries don't open their file unless needed
  if (handle->f == NULL) {
    char data_path[lsm_entry_data_path_len(handle) + 1];
    lsm_entry_data_path(data_path, handle);

    FILE *f = fopen(data_path, "rb");

    if (f == NULL) {
      return lsm_error_failed_io;
    }

    handle->f = f;
  }

  uint64_t read;

  read = fread(buf, sizeof(char), len, handle->f);

  if ((read == 0) && (ferror(handle->f) != 0)) {
    return lsm_error_failed_io;
  }

  handle->pos += read;
  *out = read;

  return lsm_error_ok;
}
